/* Copyright (C) 2009-2010 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * If the integration of miniscope code requires the use of a license
 * other than GPL, an alternative license may be arranged.
 */

#include "main.h"

#define _EXPORTING
#include "..\miniscope\device\MiniscopeDevice.h"
#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include "BusPirateScope.h"
#include "Utils.h"
#include <assert.h>
#include <fstream>
#include <json/json.h>

// a sample exported function
void DLL_EXPORT SomeFunction(const LPCSTR sometext)
{
    MessageBoxA(0, sometext, "DLL Message", MB_OK | MB_ICONINFORMATION);
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
    switch (fdwReason)
    {
    case DLL_PROCESS_ATTACH:
        // attach to process
        // return FALSE to fail DLL load
        break;

    case DLL_PROCESS_DETACH:
        // detach from process
        break;

    case DLL_THREAD_ATTACH:
        // attach to thread
        break;

    case DLL_THREAD_DETACH:
        // detach from thread
        break;
    }
    return TRUE; // succesful
}

static const struct S_DEVICE_DLL_INTERFACE dll_interface =
    {DLL_INTERFACE_MAJOR_VERSION, DLL_INTERFACE_MINOR_VERSION};

// callback ptrs
CALLBACK_LOG lpLogFn = NULL;
CALLBACK_CONNECT lpConnectFn = NULL;
CALLBACK_DATA lpDataFn = NULL;
CALLBACK_TRIGGER lpTriggerFn = NULL;


void *callbackCookie;	///< used by upper class to distinguish library instances when receiving callbacks

void __stdcall GetMiniscopeInterfaceDescription(struct S_DEVICE_DLL_INTERFACE* interf)
{
    interf->majorVersion = dll_interface.majorVersion;
    interf->minorVersion = dll_interface.minorVersion;
}

void Log(char* txt)
{
    if (lpLogFn)
        lpLogFn(callbackCookie, txt);
}

void Connect(int state, char *szMsgText)
{
    if (lpConnectFn)
        lpConnectFn(callbackCookie, state, szMsgText);
}

void Data(short* data, int size)
{
    if (lpDataFn)
        lpDataFn(callbackCookie, data, size);
}

void Trigger(void)
{
    if (lpTriggerFn)
        lpTriggerFn(callbackCookie);
}

void __stdcall SetCallbacks(void *cookie, CALLBACK_LOG lpLog, CALLBACK_CONNECT lpConnect,
                            CALLBACK_DATA lpData, CALLBACK_TRIGGER lpTrigger)
{
    assert(cookie && lpLog && lpConnect && lpData && lpTrigger);
    lpLogFn = lpLog;
    lpConnectFn = lpConnect;
    lpDataFn = lpData;
    lpTriggerFn = lpTrigger;
    callbackCookie = cookie;

    char buf[64];
    snprintf(buf, sizeof(buf), "buspirate.dll (built %s, %s) loaded!", __DATE__, __TIME__);
    buf[sizeof(buf)-1] = '\0';
    lpLogFn(callbackCookie, buf);

    busPirateScope->callbackConnect = Connect;
    busPirateScope->callbackLog = Log;
    busPirateScope->callbackData = Data;
    busPirateScope->callbackTrigger = Trigger;

    lpLogFn(callbackCookie, (char*)"Callbacks set");
}

static float fSampl[] = { 1.0/5720.0 }; // based on http://dangerousprototypes.com/forum/index.php?topic=976.0
static unsigned int iBuf[] = { 1024, 2048, 4096, 8192, 16384, 32768 };
static float fSens[] = { 6.6/1024.0 };
static enum E_TRIGGER_MODE eMode[] = { TRIGGER_MODE_AUTO, TRIGGER_MODE_NORMAL, TRIGGER_MODE_SINGLE, TRIGGER_MODE_CONTINUOUS };
static enum E_TRIGGER_TYPE eTrigger[] = { TRIGGER_SLOPE_UP, TRIGGER_SLOPE_DOWN };
static enum E_COUPLING_TYPE eCoupling[] = { COUPLING_DC };
enum { CHANNEL_CNT = 1 };
bool pChannelEnable[CHANNEL_CNT];
unsigned int pSensitivity[CHANNEL_CNT];
enum E_COUPLING_TYPE eCouplingType[CHANNEL_CNT];

void __stdcall GetDeviceCapabilities(struct S_SCOPE_CAPABILITIES **caps)
{
    lpLogFn(callbackCookie, (char*)"GetDeviceCapabilities");

    static struct S_SCOPE_CAPABILITIES capabilities =
    {
        CHANNEL_CNT,
        11,		//unsigned int iBitsPerSample;
        sizeof(fSens)/sizeof(fSens[0]), fSens,
        sizeof(fSampl)/sizeof(fSampl[0]), fSampl,
        sizeof(eMode)/sizeof(eMode[0]), eMode,
        sizeof(eTrigger)/sizeof(eTrigger[0]), eTrigger,
        sizeof(eCoupling)/sizeof(eCoupling[0]), eCoupling,
        sizeof(iBuf)/sizeof(iBuf[0]), iBuf
    };
    *caps = &capabilities;
}

void __stdcall ShowSettings(HANDLE parent)
{
    MessageBox(NULL, "Settings can be modified only by editing .cfg file. Close application before running editor.", "Device DLL", MB_ICONINFORMATION);
}

int __stdcall SetSensitivity(int iChannelId, int iId)
{
    if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
        return -1;
    return 0;
}

int __stdcall SetSamplingPeriod(int iId)
{
    if (iId < 0 || iId >= (int)(sizeof(fSampl)/sizeof(fSampl[0])))
        return -1;
    //return busPirateScope->SetSampling((int)((float)1000 * fSampl[iId] + 0.5));
    return 0;
}

int __stdcall SetTriggerSource(int iChannelId, int iId)
{
    if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
        return -1;
    if (iId < 0 || iId >= (int)(sizeof(eTrigger)/sizeof(enum E_TRIGGER_TYPE)))
        return -1;
    return busPirateScope->SetTriggerType(eTrigger[iId]);
}

int __stdcall SetTriggerLevel(int level)
{
    return busPirateScope->SetTriggerLevel(1024*level/100);
}


int __stdcall SetCouplingType(int iChannelId, int iId)
{
    if (iChannelId < 0 || iChannelId >= CHANNEL_CNT)
        return -1;
    return 0;
}

int __stdcall SetBufferSize(int iId)
{
    if (iId < 0 || iId >= (int)(sizeof(iBuf)/sizeof(iBuf[0])))
    {
        char buf[64];
        sprintf(buf, "Buffer size id = %d, out of range!", iId);
        lpLogFn(callbackCookie, buf);
        return -1;
    }
    return busPirateScope->SetBufferSize(iBuf[iId]);
}


int __stdcall Connect(void)
{
    return busPirateScope->Connect();
}

int __stdcall Disconnect(void)
{
    return busPirateScope->Disconnect();
}

int __stdcall Run(bool state)
{
    return busPirateScope->Run(state);
}

int __stdcall SetTriggerMode(int iId)
{
    if (iId < 0 || iId >= (int)(sizeof(eMode)/sizeof(eMode[0])))
        return -1;
    return busPirateScope->SetTriggerMode(eMode[iId]);
}

int __stdcall ManualTrigger(void)
{
    return busPirateScope->ManualTrigger();
}

/** \brief Calibration data for single sensitivity range
*/
static struct S_CALIBRATION
{
    int offset;
    float gain;
} calibration[sizeof(fSens)/sizeof(fSens[0])];
static bool bSettingsReaded = false;
int iComPortNumber = 3;

static int GetDefaultSettings(struct S_SCOPE_SETTINGS* settings)
{
    lpLogFn(callbackCookie, (char*)"GetDefaultSettings");
    settings->pChannelEnable = pChannelEnable;
    settings->pSensitivity = pSensitivity;
    settings->eCouplingType = eCouplingType;

    settings->pChannelEnable[0] = true;
    settings->eCouplingType[0] = COUPLING_DC;
    settings->pSensitivity[0] = 0;
    settings->iSamplingPeriod = 0;
    settings->eTriggerMode = TRIGGER_MODE_AUTO;
    settings->eTriggerType = TRIGGER_SLOPE_UP;
    settings->iTriggerLevel = 0;
    settings->iTriggerPosition = 50;
    settings->iTriggerSrcChannel = 0;
    settings->iBufferSize = iBuf[0];

    for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
    {
        calibration[i].offset = 0;
        calibration[i].gain = 1.0;
    }
    bSettingsReaded = true;
    return 0;
}

int __stdcall GetMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    lpLogFn(callbackCookie, (char*)"GetMiniscopeSettings");

    settings->pChannelEnable = pChannelEnable;
    settings->pSensitivity = pSensitivity;
    settings->eCouplingType = eCouplingType;

    settings->iTriggerSrcChannel = 0;

    std::string path = Utils::GetDllPath();
    path = Utils::ReplaceFileExtension(path, ".cfg");
    if (path == "")
    {
        lpLogFn(callbackCookie, (char*)"Could not get .cfg file path, using default settings");
        return GetDefaultSettings(settings);
    }

    Json::Value root;   // will contains the root value after parsing.
    Json::Reader reader;

    std::ifstream ifs(path.c_str());
    std::string strConfig((std::istreambuf_iterator<char>(ifs)), std::istreambuf_iterator<char>());
    ifs.close();

    bool parsingSuccessful = reader.parse( strConfig, root );
    if ( !parsingSuccessful )
        return GetDefaultSettings(settings);

    settings->pChannelEnable[0] = true;

    settings->pSensitivity[0] = root.get("Sensitivity", 0).asInt();
    if (settings->pSensitivity[0] >= sizeof(fSens)/sizeof(fSens[0]))
        settings->pSensitivity[0] = 0;

    settings->iSamplingPeriod = root.get("SamplingPeriod", 0).asInt();
    if (settings->iSamplingPeriod >= sizeof(fSampl)/sizeof(fSampl[0]))
        settings->iSamplingPeriod = 0;

	int mode = root.get("TriggerMode", TRIGGER_MODE_AUTO).asInt();
	if (Utils::in_group(mode, eMode))
		settings->eTriggerMode = (E_TRIGGER_MODE)mode;
	else
		settings->eTriggerMode = TRIGGER_MODE_AUTO;

	int trigger = root.get("TriggerType", TRIGGER_SLOPE_UP).asInt();
	if (Utils::in_group(trigger, eTrigger))
		settings->eTriggerType = (E_TRIGGER_TYPE)trigger;
	else
		settings->eTriggerType = TRIGGER_SLOPE_UP;

    settings->iTriggerLevel = root.get("TriggerLevel", 0).asInt();
    if (settings->iTriggerLevel < -100 || settings->iTriggerLevel > 100)
        settings->iTriggerLevel = 0;

    int coupling = root.get("CouplingType", COUPLING_DC).asInt();
    if (Utils::in_group(coupling, eCoupling))
        settings->eCouplingType[0] = (E_COUPLING_TYPE)coupling;
    else
        settings->eCouplingType[0] = COUPLING_DC;

    settings->iBufferSize = root.get("BufferSize", iBuf[0]).asInt();
    if (!Utils::in_group(settings->iBufferSize, iBuf))
    {
        settings->iBufferSize = iBuf[0];
    }

    //-fshort-enums option is required!
//	char buf[64];
//	sprintf(buf, "iBufferSize = %d, sizeof(*settings) = %d", settings->iBufferSize, sizeof(*settings));
//	lpLogFn(callbackCookie, buf);


    // read calibration data also
    const Json::Value rootcal = root["Calibration"];
    for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
    {
        calibration[i].offset = rootcal[i].get("Offset", 0).asInt();
        calibration[i].gain = rootcal[i].get("Gain", 1.0).asDouble();
    }

    iComPortNumber = root.get("ComPortNumber", 1).asInt();

    bSettingsReaded = true;

    return 0;
}

int __stdcall SaveMiniscopeSettings(struct S_SCOPE_SETTINGS* settings)
{
    Json::Value root;
    Json::StyledWriter writer;
    root["Sensitivity"] = settings->pSensitivity[0];
    root["SamplingPeriod"] = settings->iSamplingPeriod;
    root["TriggerType"] = settings->eTriggerType;
    root["TriggerLevel"] = settings->iTriggerLevel;
    root["CouplingType"] = settings->eCouplingType[0];
    root["BufferSize"] = settings->iBufferSize;

    root["ComPortNumber"] = iComPortNumber;

    // add calibration data also
    const Json::Value rootcal = root["Calibration"];
    for (unsigned int i=0; i<sizeof(calibration)/sizeof(calibration[0]); i++)
    {
        root["Calibration"][i]["Offset"] = calibration[i].offset;
        root["Calibration"][i]["Gain"] = calibration[i].gain;
    }

    std::string outputConfig = writer.write( root );

    std::string path = Utils::GetDllPath();
    path = Utils::ReplaceFileExtension(path, ".cfg");
    if (path == "")
    {
        lpLogFn(callbackCookie, (char*)"Could not get .cfg file path to store settings");
        return -1;
    }

    std::ofstream ofs(path.c_str());
    ofs << outputConfig;
    ofs.close();

    return 0;
}

int __stdcall StoreCalibration(int channel, int range, int offset, float gain)
{
    if (channel < 0 || channel >= CHANNEL_CNT)
        return -1;
    if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
        return -1;
    calibration[range].offset = offset;
    calibration[range].gain = gain;
    return 0;
}

int __stdcall GetCalibration(int channel, int range, int* const offset, float* const gain)
{
    lpLogFn(callbackCookie, (char*)"GetCalibration");

    if (channel < 0 || channel >= CHANNEL_CNT)
        return -1;
    if (range < 0 || range >= (int)(sizeof(fSens)/sizeof(fSens[0])))
        return -2;
    struct S_SCOPE_SETTINGS settings;
    if (bSettingsReaded == false)
    {
        int status = GetMiniscopeSettings(&settings);
        if (status)
            return status;
    }
    *offset = calibration[range].offset;
    *gain = calibration[range].gain;
    return 0;
}

